# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

import logging
import random
from math import ceil
from unittest.mock import call
from unittest.mock import patch

import pytest

from i2c_gui2.functions import bytes_to_word_list
from i2c_gui2.functions import word_list_to_bytes
from i2c_gui2.i2c_connection_helper import I2C_Connection_Helper
from i2c_gui2.i2c_messages import I2CMessages


@pytest.fixture
def i2c_ch_max_seq_byte():
    yield 8


@pytest.fixture
def i2c_ch_i2c_delay():
    yield 1000


@pytest.fixture
def i2c_ch_no_connect():
    yield True


@pytest.fixture
def i2c_ch_test(i2c_ch_max_seq_byte, i2c_ch_i2c_delay, i2c_ch_no_connect):
    yield I2C_Connection_Helper(
        max_seq_byte=i2c_ch_max_seq_byte,
        successive_i2c_delay_us=i2c_ch_i2c_delay,
        no_connect=i2c_ch_no_connect,
    )


def test_init(i2c_ch_test):
    assert not i2c_ch_test._is_connected
    assert isinstance(i2c_ch_test._logger, logging.Logger)


def test_max_seq_byte(i2c_ch_max_seq_byte, i2c_ch_test):
    assert i2c_ch_max_seq_byte == i2c_ch_test._max_seq_byte


def test_i2c_delay(i2c_ch_i2c_delay, i2c_ch_test):
    assert i2c_ch_i2c_delay == i2c_ch_test._successive_i2c_delay_us


def test_no_connect(i2c_ch_no_connect, i2c_ch_test):
    assert i2c_ch_no_connect == i2c_ch_test._no_connect


def test_logger(caplog, i2c_ch_test):
    assert isinstance(i2c_ch_test.logger, logging.Logger)
    assert i2c_ch_test.logger == i2c_ch_test._logger

    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test.logger.info("Test Debug")

    log_tuples = caplog.record_tuples
    assert len(log_tuples) == 1
    assert log_tuples[0][0] == "I2C_Log"
    assert log_tuples[0][1] == logging.INFO
    assert log_tuples[0][2] == "Test Debug"


def test_connected(i2c_ch_test):
    assert not i2c_ch_test.connected


def test_not_implemented_check_i2c_devices(i2c_ch_test):
    with pytest.raises(Exception) as e_info:
        i2c_ch_test._check_i2c_device(0x21)
    assert e_info.match(r"^Derived classes must implement the individual device access methods:")


def test_not_implemented_write_i2c_device_memory(i2c_ch_test):
    with pytest.raises(Exception) as e_info:
        i2c_ch_test._write_i2c_device_memory(0x21, 10, [1, 2])
    assert e_info.match(r"^Derived classes must implement the individual device access methods:")


def test_not_implemented_read_i2c_device_memory(i2c_ch_test):
    with pytest.raises(Exception) as e_info:
        i2c_ch_test._read_i2c_device_memory(0x21, 10, 5)
    assert e_info.match(r"^Derived classes must implement the individual device access methods:")


def test_not_implemented_direct_i2c(i2c_ch_test):
    with pytest.raises(Exception) as e_info:
        i2c_ch_test._direct_i2c([I2CMessages.START, I2CMessages.STOP])
    assert e_info.match(r"^Derived classes must implement the individual device access methods:")


@pytest.fixture()
def connect_return_value():
    yield None


@pytest.fixture()
def fake_connect(monkeypatch, connect_return_value):
    def replace():
        return connect_return_value

    monkeypatch.setattr(I2C_Connection_Helper, "_check_i2c_device", lambda *args, **kwargs: replace())


@pytest.mark.parametrize('connect_return_value', [True, False])
def test_fake_connect_monkeypatch(connect_return_value, fake_connect, i2c_ch_test):
    assert i2c_ch_test._check_i2c_device(0x21) == connect_return_value


@pytest.mark.parametrize('i2c_ch_no_connect', [True, False])
@pytest.mark.parametrize('connect_return_value', [True, False])
def test_check_i2c_device(caplog, connect_return_value, fake_connect, i2c_ch_no_connect, i2c_ch_test):
    caplog.set_level(logging.DEBUG, "I2C_Log")

    assert i2c_ch_test._check_i2c_device(0x21) == connect_return_value  # Sanity check the monkeypatching works

    i2c_ch_test._is_connected = not i2c_ch_no_connect

    found = i2c_ch_test.check_i2c_device(0x21)

    if i2c_ch_no_connect:
        assert not found
    else:
        if connect_return_value is False:
            assert not found
        else:
            assert found

    log_tuples = caplog.record_tuples
    assert len(log_tuples) == 2
    assert log_tuples[0][0] == "I2C_Log"
    assert log_tuples[1][0] == "I2C_Log"
    assert log_tuples[0][1] == logging.INFO
    assert log_tuples[1][1] == logging.INFO
    # assert log_tuples[0][2] == "Test Debug"


def test_read_device_memory_not_connected(i2c_ch_test):
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.read_device_memory(0x21, 0x00, 1)
    assert e_info.match(r"^You must first connect to a device before trying to read registers from it")


def test_write_device_memory_not_connected(i2c_ch_test):
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.write_device_memory(0x21, 0x00, [0x54])
    assert e_info.match(r"^You must first connect to a device before trying to write registers to it")


def test_read_device_memory_invalid_device_address(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.read_device_memory(0x80, 0x00, 1)
    assert e_info.match(r"^Invalid I2C address received: 0x80")


def test_write_device_memory_invalid_device_address(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.write_device_memory(0x80, 0x00, [0x54])
    assert e_info.match(r"^Invalid I2C address received: 0x80")


def test_read_device_memory_invalid_address_endianness(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.read_device_memory(0x21, 0x00, 1, address_endianness='blabla')
    assert e_info.match(r"^A wrong address endianness was set: blabla")


def test_write_device_memory_invalid_address_endianness(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.write_device_memory(0x21, 0x00, [0x54], address_endianness='blabla')
    assert e_info.match(r"^A wrong address endianness was set: blabla")


def test_read_device_memory_invalid_word_endianness(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.read_device_memory(0x21, 0x00, 1, word_endianness='blabla')
    assert e_info.match(r"^A wrong word endianness was set: blabla")


def test_write_device_memory_invalid_word_endianness(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.write_device_memory(0x21, 0x00, [0x54], word_endianness='blabla')
    assert e_info.match(r"^A wrong word endianness was set: blabla")


def test_read_device_memory_invalid_read_type(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.read_device_memory(0x21, 0x00, 1, read_type='blabla')
    assert e_info.match(r"^A wrong read type was set: blabla")


def test_write_device_memory_invalid_read_type(i2c_ch_test):
    i2c_ch_test._is_connected = True
    with pytest.raises(Exception) as e_info:
        i2c_ch_test.write_device_memory(0x21, 0x00, 1, write_type='blabla')
    assert e_info.match(r"^A wrong write type was set: blabla")


@pytest.mark.parametrize('words', [1, 2, 4])
@pytest.mark.parametrize('bitlength', [8, 16])
@pytest.mark.parametrize('endianness', ['big', 'little'])
def test_read_device_memory_no_connect(caplog, i2c_ch_test, words, bitlength, endianness):
    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test._is_connected = True

    word_list = i2c_ch_test.read_device_memory(0x21, 0x00, words, word_bitlength=bitlength, word_endianness=endianness)

    log_tuples = caplog.record_tuples
    assert len(log_tuples) == 2
    assert log_tuples[0][0] == "I2C_Log"
    assert log_tuples[1][0] == "I2C_Log"
    assert log_tuples[0][1] == logging.INFO
    assert log_tuples[1][1] == logging.DEBUG
    assert "Software emulation (no connect) is enabled, so returning dummy values:" in log_tuples[1][2]

    assert len(word_list) == words
    if words == 1:
        assert "Reading the register" in log_tuples[0][2]
        assert word_list[0] == 42
    else:
        assert "Reading a register block with size" in log_tuples[0][2]
        if bitlength == 8:
            for i in range(words):
                assert word_list[i] == i
        else:
            byte_list = [i for i in range(words * ceil(bitlength / 8))]
            compare_list = bytes_to_word_list(byte_list, ceil(bitlength / 8), endianness)
            assert word_list == compare_list


@pytest.mark.parametrize('words', [1, 2, 4])
@pytest.mark.parametrize('bitlength', [8, 16])
@pytest.mark.parametrize('endianness', ['big', 'little'])
def test_write_device_memory_no_connect(caplog, i2c_ch_test, words, bitlength, endianness):
    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test._is_connected = True

    word_list = [i for i in range(words)]

    i2c_ch_test.write_device_memory(0x21, 0x00, word_list, word_bitlength=bitlength, word_endianness=endianness)

    log_tuples = caplog.record_tuples
    assert len(log_tuples) == 2
    assert log_tuples[0][0] == "I2C_Log"
    assert log_tuples[1][0] == "I2C_Log"
    assert log_tuples[0][1] == logging.INFO
    assert log_tuples[1][1] == logging.DEBUG
    assert log_tuples[1][2] == "Software emulation (no connect) is enabled, so no write action is taken."

    assert len(word_list) == words
    if words == 1:
        assert "Writing the value" in log_tuples[0][2]
    else:
        assert "Writing a register block with size" in log_tuples[0][2]


@pytest.mark.parametrize('i2c_ch_max_seq_byte', [None])
@pytest.mark.parametrize('i2c_ch_no_connect', [False])
@pytest.mark.parametrize('words', [1, 2])
@pytest.mark.parametrize('bitlength', [8, 16])
@pytest.mark.parametrize('endianness', ['big', 'little'])
def test_read_device_memory_no_max_seq_byte(caplog, i2c_ch_test, words, bitlength, endianness):
    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test._is_connected = True

    with patch('i2c_gui2.i2c_connection_helper.I2C_Connection_Helper._read_i2c_device_memory') as function:
        ret_list = [random.getrandbits(8) for _ in range(words * ceil(bitlength / 8))]
        function.return_value = ret_list

        word_list = i2c_ch_test.read_device_memory(0x21, 0x00, words, word_bitlength=bitlength, word_endianness=endianness)

        function.assert_called_once()
        function.assert_has_calls([call(0x21, 0x00, words * ceil(bitlength / 8), read_type="Normal", address_bitlength=8)])

        log_tuples = caplog.record_tuples
        assert len(log_tuples) == 2
        assert log_tuples[0][0] == "I2C_Log"
        assert log_tuples[1][0] == "I2C_Log"
        assert log_tuples[0][1] == logging.INFO
        assert log_tuples[1][1] == logging.DEBUG
        assert "Got data:" in log_tuples[1][2]

        assert len(word_list) == words

        compare_list = bytes_to_word_list(ret_list, ceil(bitlength / 8), endianness)
        assert word_list == compare_list

        if words == 1:
            assert "Reading the register" in log_tuples[0][2]
        else:
            assert "Reading a register block with size" in log_tuples[0][2]


@pytest.mark.parametrize('i2c_ch_max_seq_byte', [None])
@pytest.mark.parametrize('i2c_ch_no_connect', [False])
@pytest.mark.parametrize('words', [1, 2])
@pytest.mark.parametrize('bitlength', [8, 16])
@pytest.mark.parametrize('endianness', ['big', 'little'])
def test_write_device_memory_no_max_seq_byte(caplog, i2c_ch_test, words, bitlength, endianness):
    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test._is_connected = True

    word_list = [random.getrandbits(bitlength) for _ in range(words)]
    byte_list = word_list_to_bytes(word_list, ceil(bitlength / 8), endianness)

    with patch('i2c_gui2.i2c_connection_helper.I2C_Connection_Helper._write_i2c_device_memory') as function:
        i2c_ch_test.write_device_memory(0x21, 0x00, word_list, word_bitlength=bitlength, word_endianness=endianness)

        function.assert_called_once()
        function.assert_has_calls([call(0x21, 0x00, byte_list, write_type="Normal", address_bitlength=8)])

        log_tuples = caplog.record_tuples
        assert len(log_tuples) == 2
        assert log_tuples[0][0] == "I2C_Log"
        assert log_tuples[1][0] == "I2C_Log"
        assert log_tuples[0][1] == logging.INFO
        assert log_tuples[1][1] == logging.DEBUG
        assert log_tuples[1][2] == "Writing the full block at once."

        if words == 1:
            assert "Writing the value" in log_tuples[0][2]
        else:
            assert "Writing a register block with size" in log_tuples[0][2]


@pytest.mark.parametrize('i2c_ch_max_seq_byte', [1, 2, 8])
@pytest.mark.parametrize('i2c_ch_no_connect', [False])
@pytest.mark.parametrize('words', [1, 2, 8])
@pytest.mark.parametrize('bitlength', [8, 16])
@pytest.mark.parametrize('endianness', ['big', 'little'])
def test_read_device_memory(caplog, i2c_ch_test, words, bitlength, endianness, i2c_ch_max_seq_byte):
    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test._is_connected = True

    with patch('i2c_gui2.i2c_connection_helper.I2C_Connection_Helper._read_i2c_device_memory') as function:
        ret_list = [random.getrandbits(8) for _ in range(min(words * ceil(bitlength / 8), i2c_ch_max_seq_byte))]
        function.return_value = ret_list

        if ceil(bitlength / 8) > i2c_ch_max_seq_byte:
            with pytest.raises(Exception) as e_info:
                word_list = i2c_ch_test.read_device_memory(0x21, 0x00, words, word_bitlength=bitlength, word_endianness=endianness)
            assert e_info.match(r"^The word length is too big for the maximum number of bytes")

            function.assert_not_called()
        else:
            word_list = i2c_ch_test.read_device_memory(0x21, 0x00, words, word_bitlength=bitlength, word_endianness=endianness)

            num_calls = ceil(words * ceil(bitlength / 8) / i2c_ch_max_seq_byte)
            addr_step = i2c_ch_max_seq_byte / (bitlength / 8)

            function.assert_has_calls(
                [
                    call(
                        0x21,
                        int(i * addr_step),
                        min(words * ceil(bitlength / 8), i2c_ch_max_seq_byte),
                        read_type="Normal",
                        address_bitlength=8,
                    )
                    for i in range(num_calls)
                ]
            )

            log_tuples = caplog.record_tuples
            assert len(log_tuples) == 2 * num_calls + 2
            for i in range(2 * num_calls + 2):
                assert log_tuples[i][0] == "I2C_Log"
                if i == 0:
                    assert log_tuples[i][1] == logging.INFO
                else:
                    assert log_tuples[i][1] == logging.DEBUG
                if i == 0:
                    continue
                elif i == 1:
                    assert f"Breaking the read into {num_calls}" in log_tuples[i][2]
                else:
                    if i % 2 == 0:
                        assert f"Read operation {int(i/2 - 1)}" in log_tuples[i][2]
                    else:
                        assert "Got data:" in log_tuples[i][2]

            assert len(word_list) == words

            compare_list = []
            for i in range(num_calls):
                compare_list += ret_list
            compare_list = bytes_to_word_list(compare_list, ceil(bitlength / 8), endianness)
            assert word_list == compare_list

            if words == 1:
                assert "Reading the register" in log_tuples[0][2]
            else:
                assert "Reading a register block with size" in log_tuples[0][2]


@pytest.mark.parametrize('i2c_ch_max_seq_byte', [1, 2, 8])
@pytest.mark.parametrize('i2c_ch_no_connect', [False])
@pytest.mark.parametrize('words', [1, 2, 8])
@pytest.mark.parametrize('bitlength', [8, 16])
@pytest.mark.parametrize('endianness', ['big', 'little'])
def test_write_device_memory(caplog, i2c_ch_test, words, bitlength, endianness, i2c_ch_max_seq_byte):
    caplog.set_level(logging.DEBUG, "I2C_Log")
    i2c_ch_test._is_connected = True

    word_list = [random.getrandbits(bitlength) for _ in range(words)]
    byte_list = word_list_to_bytes(word_list, ceil(bitlength / 8), endianness)

    with patch('i2c_gui2.i2c_connection_helper.I2C_Connection_Helper._write_i2c_device_memory') as function:

        if ceil(bitlength / 8) > i2c_ch_max_seq_byte:
            with pytest.raises(Exception) as e_info:
                i2c_ch_test.write_device_memory(0x21, 0x00, word_list, word_bitlength=bitlength, word_endianness=endianness)
            assert e_info.match(r"^The word length is too big for the maximum number of bytes")

            function.assert_not_called()
        else:
            i2c_ch_test.write_device_memory(0x21, 0x00, word_list, word_bitlength=bitlength, word_endianness=endianness)

            num_calls = ceil(words * ceil(bitlength / 8) / i2c_ch_max_seq_byte)
            addr_step = i2c_ch_max_seq_byte / (bitlength / 8)

            function.assert_has_calls(
                [
                    call(
                        0x21,
                        int(i * addr_step),
                        byte_list[
                            i * i2c_ch_max_seq_byte : i * i2c_ch_max_seq_byte + min(words * ceil(bitlength / 8), i2c_ch_max_seq_byte)
                        ],
                        write_type="Normal",
                        address_bitlength=8,
                    )
                    for i in range(num_calls)
                ]
            )

            log_tuples = caplog.record_tuples
            assert len(log_tuples) == 2 * num_calls + 2
            for i in range(2 * num_calls + 2):
                assert log_tuples[i][0] == "I2C_Log"
                if i == 0:
                    assert log_tuples[i][1] == logging.INFO
                else:
                    assert log_tuples[i][1] == logging.DEBUG
                if i == 0:
                    continue
                elif i == 1:
                    assert f"Breaking the write into {num_calls}" in log_tuples[i][2]
                else:
                    if i % 2 == 0:
                        assert f"Write operation {int(i/2 - 1)}" in log_tuples[i][2]
                    else:
                        assert "Current block:" in log_tuples[i][2]

            if words == 1:
                assert "Writing the value" in log_tuples[0][2]
            else:
                assert "Writing a register block with size" in log_tuples[0][2]



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################
"""The i2c_connection_helper module

Contains the I2C_Connection_Helper class, which is only a base class from which
derived classes which implement I2C communication should inherit from.

"""

from __future__ import annotations

import logging
from math import ceil
from math import floor
from time import sleep
from time import time_ns

from .functions import address_to_phys
from .functions import bytes_to_word_list
from .functions import valid_i2c_address
from .functions import word_list_to_bytes
from .i2c_messages import I2CMessages

valid_endianness = ['little', 'big']
valid_read_type = ['Normal', 'Repeated Start']
valid_write_type = ['Normal']


class I2C_Connection_Helper:
    """Base Class to handle an I2C Connection

    This is a base class from which derived classes should inherit and implement
    the necessary methods.

    This base class sets up many of the internal features of an I2C connection,
    such as an internal log for logging I2C activity.

    Parameters
    ----------
    max_seq_byte
        The maximum number of bytes which can be transmitted in a single I2C command

    successive_i2c_delay_us
        The minimum delay in microseconds (us) between successive I2C commands

    no_connect
        Mostly used for debugging, if set to True, no physical cconnection will
        actually be attempted, but default values will be returned as I2C responses

    """

    def __init__(
        self,
        max_seq_byte: int,
        successive_i2c_delay_us: int = 10000,
        no_connect: bool = False,
    ):
        self._max_seq_byte = max_seq_byte

        self._no_connect = no_connect

        self._successive_i2c_delay_us = successive_i2c_delay_us

        self._logger = logging.getLogger("I2C_Log")
        self._logger.setLevel(logging.NOTSET)

        self._is_connected = False

        self._lastI2COperation = time_ns()

    @property
    def logger(self):
        """The logger property getter method

        This method returns the internal logger of the I2C Connection object

        Returns
        -------
        logging.Logger
            The logger
        """
        return self._logger

    @property
    def connected(self):
        """The connected property getter method

        This method returns the connection status of the I2C connection

        Returns
        -------
        bool
            The state of the connection
        """
        return self._is_connected

    def _check_i2c_device(self, device_address: int) -> bool:
        """The internal method to check if an i2c device with the given address is connected

        This method must be implemented by the derived classes

        Parameters
        ----------
        device_address
            The I2C address of the device to check. It must be a 7-bit address as per the I2C standard

        Raises
        ------
        RuntimeError
            If the derived class has not implemented this method

        Returns
        -------
        bool
            The presence or absence of the device with the `device_address`
        """
        raise RuntimeError("Derived classes must implement the individual device access methods: _check_i2c_device")

    def _write_i2c_device_memory(
        self,
        device_address: int,
        word_address: int,
        byte_data: list[int],
        write_type: str = 'Normal',
        address_bitlength: int = 8,
    ):
        """The internal method to write byte data to an i2c device with the given address.

        This method must be implemented by the derived classes.

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        word_address
            The address of the first byte to be written to. The address should have its endianness correctly
            set such that the MSB of `word_address` is the first byte sent on the I2C lines.

        byte_data
            The byte data to be written to the I2C device. The data is written to the I2C bus in the order
            given, so the endianness of the data must be correctly set, assuming the device contains registers
            larger than 8 bits.

        write_type
            The type of protocol used for the actual writing procedure to the device.

        address_bitlength
            The length in bits of the address

        Raises
        ------
        RuntimeError
            If the derived class has not implemented this method
        """
        raise RuntimeError("Derived classes must implement the individual device access methods: _write_i2c_device_memory")

    def _read_i2c_device_memory(
        self,
        device_address: int,
        word_address: int,
        byte_count: int,
        read_type: str = 'Normal',
        address_bitlength: int = 8,
    ) -> list[int]:
        """The internal method to read byte data from an i2c device with the given address.

        This method must be implemented by the derived classes.

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        word_address
            The address of the first byte to be read from. The address should have its endianness correctly
            set such that the MSB of `word_address` is the first byte sent on the I2C lines.

        byte_count
            The number of bytes to be read from the I2C device. If multybyte registers are to be read, the
            number of bytes should be given, not the number of words.

        read_type
            The type of protocol used for the actual reading procedure from the device. Supported protocols
            are "Normal" and "Repeated Start". The Repeated Start protocol is implemented by the AD5593R chip.

        address_bitlength
            The length in bits of the address

        Raises
        ------
        RuntimeError
            If the derived class has not implemented this method

        Returns
        -------
        list[int]
            The list of bytes in the order presented on the I2C bus. If multibyte registers are being read, then
            the words must be correctly put back together by taking into account the data endianness sent by the
            I2C device.
        """
        raise RuntimeError("Derived classes must implement the individual device access methods: _read_i2c_device_memory")

    def _direct_i2c(self, commands: list[I2CMessages]) -> list[int]:
        """The internal method to send arbitrary I2C messages to the I2C bus.

        This method must be implemented by the derived classes.

        Parameters
        ----------
        commands
            The I2C commands to be sent to the I2C bus in the order they are to be sent.

        Raises
        ------
        RuntimeError
            If the derived class has not implemented this method

        Returns
        -------
        list[int]
            The list of bytes returned to the I2C Bus in the order presented on the I2C bus.
        """
        raise RuntimeError("Derived classes must implement the individual device access methods: _direct_i2c")

    def check_i2c_device(self, device_address: int) -> bool:
        """The user method to check if a device with the `device_address` is connected to the I2C bus.

        This method makes use of the internal method _check_i2c_device

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        Raises
        ------
        RuntimeError
            If there is any problem during runtime

        Returns
        -------
        bool
            The presence or absence of the device with the `device_address`
        """
        self._logger.info("Trying to find the I2C device with address 0x{:02x}".format(device_address))

        if not self._is_connected or self._no_connect:
            self._logger.info("The I2C device is not connected or you are using software emulated mode.")
            return False

        now = time_ns()
        if now - self._lastI2COperation < self._successive_i2c_delay_us * 1000:
            sleep(self._successive_i2c_delay_us * 10**-6)
        self._lastI2COperation = now

        if not self._check_i2c_device(device_address):
            self._logger.info("The I2C device 0x{:02x} can not be found.".format(device_address))
            return False

        self._logger.info("The I2C device 0x{:02x} was found.".format(device_address))
        return True

    def read_device_memory(
        self,
        device_address: int,
        word_address: int,
        word_count: int = 1,
        address_bitlength: int = 8,
        address_endianness: str = 'big',
        word_bitlength: int = 8,
        word_endianness: str = 'big',
        read_type: str = 'Normal',
    ) -> list[int]:
        """The user method to read register data from a device on the I2C bus with the given `device_address`.

        This method makes use of the internal method _read_i2c_device_memory

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        word_address
            The address of the first byte to be read from.

        word_count
            The number of register words to be read from the I2C device.

        address_bitlength
            The bit length of the address, typical values are 8 and 16.

        address_endianness
            The endianness of the address as presented on the I2C bus. This parameter does not make any
            difference for 8-bit addresses.

        word_bitlength
            The bit length of the register words, typical values are 8 and 16.

        word_endianness
            The endianness of the register words as presented on the I2C bus. This parameter does not make any
            difference for 8-bit register words.

        read_type
            The type of protocol used for the actual reading procedure from the device. Supported protocols
            are "Normal" and "Repeated Start". The Repeated Start protocol is implemented by the AD5593R chip.

        Raises
        ------
        RuntimeError
            If there is an issue identified during runtime

        Returns
        -------
        list[int]
            The list of word in order. The words have been put together according to the endianness options set.
        """
        if not self._is_connected:
            raise RuntimeError("You must first connect to a device before trying to read registers from it")

        if not valid_i2c_address(device_address):
            raise RuntimeError("Invalid I2C address received: {:#04x}".format(device_address))

        if address_endianness not in valid_endianness:
            raise RuntimeError(f"A wrong address endianness was set: {address_endianness}")

        if word_endianness not in valid_endianness:
            raise RuntimeError(f"A wrong word endianness was set: {word_endianness}")

        if read_type not in valid_read_type:
            raise RuntimeError(f"A wrong read type was set: {read_type}")

        word_bytes = ceil(word_bitlength / 8)

        address_chars = ceil(address_bitlength / 4)

        if word_count == 1:
            self._logger.info(
                (f"Reading the register {{:#0{address_chars+2}x}} of the I2C device with address {{:#04x}}:").format(
                    word_address, device_address
                )
            )
        else:
            self._logger.info(
                (
                    f"Reading a register block with size {{}} starting at register {{:#0{address_chars+2}x}} of"
                    f" the I2C device with address {{:#04x}}:"
                ).format(word_count, word_address, device_address)
            )

        byte_data = []
        if self._no_connect:
            if word_count == 1:
                if word_bytes == 1:
                    byte_data = [42]
                else:
                    byte_data = [0 for _ in range(word_bytes - 1)]
                    if word_endianness == 'big':
                        byte_data += [42]
                    else:  # if word_endianness == 'little':
                        byte_data = [42] + byte_data
            else:
                byte_data = [i for i in range(word_count * word_bytes)]
            self._logger.debug("Software emulation (no connect) is enabled, so returning dummy values: {}".format(repr(byte_data)))
        elif self._max_seq_byte is None:
            word_address = address_to_phys(word_address, address_bitlength, address_endianness)
            now = time_ns()
            if now - self._lastI2COperation < self._successive_i2c_delay_us * 1000:
                sleep(self._successive_i2c_delay_us * 10**-6)
            byte_data = self._read_i2c_device_memory(
                device_address, word_address, word_count * word_bytes, read_type=read_type, address_bitlength=address_bitlength
            )
            self._lastI2COperation = now
            self._logger.debug("Got data: {}".format(repr(byte_data)))
        else:
            byte_data = []
            words_per_call = floor(self._max_seq_byte / word_bytes)
            if words_per_call == 0:
                raise RuntimeError(
                    "The word length is too big for the maximum number of bytes in a single call, it is impossible"
                    " to read data in these conditions"
                )
            sequential_calls = ceil(word_count / words_per_call)
            self._logger.debug("Breaking the read into {} individual reads of {} words".format(sequential_calls, words_per_call))

            for i in range(sequential_calls):
                # Add here the possibility to call an external update function (for progress bars in GUI for instance)

                this_block_address = word_address + i * words_per_call
                this_block_words = min(words_per_call, word_count - i * words_per_call)
                bytes_to_read = this_block_words * word_bytes

                self._logger.debug(
                    (f"Read operation {{}}: reading {{}} words starting from {{:#0{address_chars+2}x}}").format(
                        i, this_block_words, this_block_address
                    )
                )

                this_block_address = address_to_phys(this_block_address, address_bitlength, address_endianness)
                now = time_ns()
                if now - self._lastI2COperation < self._successive_i2c_delay_us * 1000:
                    sleep(self._successive_i2c_delay_us * 10**-6)
                this_data = self._read_i2c_device_memory(
                    device_address, this_block_address, bytes_to_read, read_type=read_type, address_bitlength=address_bitlength
                )
                self._lastI2COperation = now
                self._logger.debug("Got data: {}".format(repr(this_data)))

                byte_data += this_data

            # Clear the progress from the function above

        # Merge byte data back into words
        return bytes_to_word_list(byte_data, word_bytes, word_endianness)

    def write_device_memory(
        self,
        device_address: int,
        word_address: int,
        data: list[int],
        address_bitlength: int = 8,
        address_endianness: str = 'big',
        word_bitlength: int = 8,
        word_endianness: str = 'big',
        write_type: str = 'Normal',
    ):
        """The user method to write register data to a device on the I2C bus with the given `device_address`.

        This method makes use of the internal method _write_i2c_device_memory

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        word_address
            The address of the first byte to be written to.

        data
            The register word data to be written to the I2C device.

        address_bitlength
            The bit length of the address, typical values are 8 and 16.

        address_endianness
            The endianness of the address as presented on the I2C bus. This parameter does not make any
            difference for 8-bit addresses.

        word_bitlength
            The bit length of the register words, typical values are 8 and 16.

        word_endianness
            The endianness of the register words as presented on the I2C bus. This parameter does not make any
            difference for 8-bit register words.

        write_type
            The type of protocol used for the actual writing procedure to the device.

        Raises
        ------
        RuntimeError
            If there is an issue identified during runtime
        """
        if not self._is_connected:
            raise RuntimeError("You must first connect to a device before trying to write registers to it")

        if not valid_i2c_address(device_address):
            raise RuntimeError("Invalid I2C address received: {:#04x}".format(device_address))

        if address_endianness not in valid_endianness:
            raise RuntimeError(f"A wrong address endianness was set: {address_endianness}")

        if word_endianness not in valid_endianness:
            raise RuntimeError(f"A wrong word endianness was set: {word_endianness}")

        if write_type not in valid_read_type:
            raise RuntimeError(f"A wrong write type was set: {write_type}")

        address_chars = ceil(address_bitlength / 4)
        word_chars = ceil(word_bitlength / 4)
        word_bytes = ceil(word_bitlength / 8)
        word_count = len(data)

        if word_count == 1:
            self._logger.info(
                (
                    f"Writing the value {{:#0{word_chars+2}x}} to the register {{:#0{address_chars+2}x}} of"
                    f" the I2C device with address {{:#04x}}:"
                ).format(data[0], word_address, device_address)
            )
        else:
            self._logger.info(
                (
                    f"Writing a register block with size {{}} starting at register {{:#0{address_chars+2}x}} of"
                    f" the I2C device with address {{:#04x}}. Writing the value array: {{}}"
                ).format(word_count, word_address, device_address, repr(data))
            )

        if self._no_connect:
            self._logger.debug("Software emulation (no connect) is enabled, so no write action is taken.")
        elif self._max_seq_byte is None:
            self._logger.debug("Writing the full block at once.")
            word_address = address_to_phys(word_address, address_bitlength, address_endianness)
            byte_data = word_list_to_bytes(data, word_bytes, word_endianness)
            now = time_ns()
            if now - self._lastI2COperation < self._successive_i2c_delay_us * 1000:
                sleep(self._successive_i2c_delay_us * 10**-6)
            self._write_i2c_device_memory(
                device_address, word_address, byte_data, write_type=write_type, address_bitlength=address_bitlength
            )
            self._lastI2COperation = now
        else:
            words_per_call = floor(self._max_seq_byte / word_bytes)
            if words_per_call == 0:
                raise RuntimeError(
                    "The word length is too big for the maximum number of bytes in a single call, it is impossible to"
                    " write data in these conditions"
                )
            sequential_calls = ceil(word_count / words_per_call)
            self._logger.debug("Breaking the write into {} individual writes of {} words".format(sequential_calls, words_per_call))

            for i in range(sequential_calls):
                # Add here the possibility to call an external update function (for progress bars in GUI for instance)

                this_block_address = word_address + i * words_per_call
                this_block_words = min(words_per_call, word_count - i * words_per_call)
                bytes_to_write = this_block_words * word_bytes
                self._logger.debug(
                    (f"Write operation {{}}: writing {{}} words starting from {{:#0{address_chars+2}x}}").format(
                        i, bytes_to_write, this_block_address
                    )
                )

                this_data = data[i * words_per_call : i * words_per_call + this_block_words]
                self._logger.debug("Current block: {}".format(repr(this_data)))

                this_block_address = address_to_phys(this_block_address, address_bitlength, address_endianness)
                this_byte_data = word_list_to_bytes(this_data, word_bytes, word_endianness)
                now = time_ns()
                if now - self._lastI2COperation < self._successive_i2c_delay_us * 1000:
                    sleep(self._successive_i2c_delay_us * 10**-6)
                self._write_i2c_device_memory(
                    device_address, this_block_address, this_byte_data, write_type=write_type, address_bitlength=address_bitlength
                )
                self._lastI2COperation = now

            # Clear the progress from the function above



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

import io
import logging
import tkinter as tk
import tkinter.ttk as ttk  # For themed widgets (gives a more native visual to the elements)

from .base_gui import Base_GUI


class Logging_Helper:
    _parent: Base_GUI
    _log_levels = {
        "Detailed Trace": logging.DETAILED_TRACE,  # (Custom)
        "Trace": logging.TRACE,  # (Custom)
        "Debug": logging.DEBUG,
        "Info": logging.INFO,
        "Warning": logging.WARNING,
        "Error": logging.ERROR,
        "Critical": logging.CRITICAL,
    }
    _default_log_level = "Info"

    def __init__(
        self,
        parent: Base_GUI,
        logger: logging.Logger,
    ):
        self._parent = parent
        self._logger = logger

        self._do_logging = False
        self._logger.disabled = True

        self._stream = io.StringIO()
        self._stream_handler = logging.StreamHandler(self._stream)
        self._stream_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s:%(name)s:%(message)s'))
        self._logger.handlers.clear()

        self._logging_window_status_var = tk.StringVar()
        self._logging_window_status_var.set("Logging Disabled")

        self._autorefresh_var = tk.BooleanVar(value=False)
        self._log_to_terminal_var = tk.BooleanVar(value=self._logger.propagate)
        self._log_level_var = tk.StringVar()
        self._log_level_var.trace_add('write', self._update_log_level)
        self._log_level_var.set(self._default_log_level)

    def _update_log_level(self, var=None, index=None, mode=None):
        self._logger.detailed_trace("Logging_Helper._update_log_level: Entered")

        new_level = self._log_level_var.get()
        if new_level not in self._log_levels:
            self._log_level_var.set(self._default_log_level)
            new_level = self._default_log_level

        numeric_level = self._log_levels[new_level]

        self._logger.setLevel(numeric_level)

    @property
    def is_logging(self):
        return self._do_logging

    @is_logging.setter
    def is_logging(self, value):
        self._logger.detailed_trace(f"Logging_Helper.is_logging setter: Entered with value {value}")

        if value not in [True, False]:
            raise TypeError("Logging can only be true or false")

        self._do_logging = value

        if self._do_logging:
            if self._stream_handler not in self._logger.handlers:
                self._logger.addHandler(self._stream_handler)
            # self._log_level_option.config(state = 'disabled')
            self._logger.disabled = False
            self._logging_window_status_var.set("Logging Enabled")
        else:
            if self._stream_handler in self._logger.handlers:
                self._logger.removeHandler(self._stream_handler)
            # self._log_level_option.config(state = 'normal')
            self._logger.disabled = True
            self._logging_window_status_var.set("Logging Disabled")

    def get_log(self):
        return self._stream.getvalue()

    def display_logging(self):
        if hasattr(self, "_logging_window"):
            self._logger.info("Logging window already open")
            self._logging_window.focus()
            return

        self._logging_window = tk.Toplevel(self._parent._root)
        self._logging_window.title(self._parent._title + " - Event Log")
        self._logging_window.protocol('WM_DELETE_WINDOW', self.close_logging_window)
        self._logging_window.columnconfigure(100, weight=1)
        self._logging_window.rowconfigure(100, weight=1)

        self._frame = ttk.Frame(self._logging_window, padding="5 5 5 5")
        self._frame.grid(column=100, row=100, sticky=(tk.N, tk.W, tk.E, tk.S))
        self._frame.columnconfigure(100, weight=1)
        self._frame.rowconfigure(100, weight=1)

        self._text_frame = ttk.Frame(self._frame)
        self._text_frame.grid(column=100, row=100, sticky=(tk.N, tk.W, tk.E, tk.S))
        self._text_frame.columnconfigure(100, weight=1)
        self._text_frame.rowconfigure(100, weight=1)

        self._text_display = tk.Text(self._text_frame, state='disabled', width=150, wrap='none')
        self._text_display.grid(column=100, row=100, sticky=(tk.N, tk.W, tk.E, tk.S))

        self._scrollbar = ttk.Scrollbar(self._text_frame, command=self._text_display.yview)
        self._scrollbar.grid(row=100, column=101, sticky='nsew')
        self._text_display.config(yscrollcommand=self._scrollbar.set)

        self._control_frame = ttk.Frame(self._frame)
        self._control_frame.grid(column=100, row=200, sticky=(tk.N, tk.W, tk.E, tk.S))

        self._toggle_logging_button = ttk.Button(self._control_frame, text="Enable Logging", command=self.toggle_logging)
        self._toggle_logging_button.grid(column=100, row=100, sticky=(tk.W, tk.E), padx=(0, 5))

        self._clear_logging_button = ttk.Button(self._control_frame, text="Clear Log", command=self.clear_log)
        self._clear_logging_button.grid(column=110, row=100, sticky=(tk.W, tk.E), padx=(0, 5))

        self._log_level_option = ttk.OptionMenu(
            self._control_frame, self._log_level_var, self._log_level_var.get(), *self._log_levels.keys()
        )
        self._log_level_option.grid(column=120, row=100, sticky=(tk.W, tk.E), padx=(0, 5))

        self._logging_status_label = ttk.Label(self._control_frame, textvariable=self._logging_window_status_var)
        self._logging_status_label.grid(column=200, row=100, sticky=(tk.W, tk.E), padx=(0, 30))
        self._control_frame.columnconfigure(200, weight=1)

        self._log_to_terminal_check = ttk.Checkbutton(
            self._control_frame, text="Log to Terminal", variable=self._log_to_terminal_var, command=self.toggle_log_to_terminal
        )
        self._log_to_terminal_check.grid(column=300, row=100, sticky=(tk.W, tk.E), padx=(0, 10))

        self._autorefresh_check = ttk.Checkbutton(
            self._control_frame, text="Auto-refresh", variable=self._autorefresh_var, command=self.toggle_autorefresh
        )
        self._autorefresh_check.grid(column=400, row=100, sticky=(tk.W, tk.E), padx=(0, 10))

        self._refresh_button = ttk.Button(self._control_frame, text="Refresh", command=self.refresh_logging)
        self._refresh_button.grid(column=500, row=100, sticky=(tk.W, tk.E))

        self._logging_window.update()
        self._logging_window.minsize(self._logging_window.winfo_width(), self._logging_window.winfo_height())

    def close_logging_window(self):
        if not hasattr(self, "_logging_window"):
            self._logger.info("Logging window does not exist")
            return

        self.is_logging = False

        self._logging_window.destroy()
        del self._logging_window

    def clear_log(self):
        self._stream.truncate(0)
        self._stream.seek(0)
        self.refresh_logging()

    def toggle_logging(self):
        self.is_logging = not self.is_logging

        button_text = "Enable Logging"
        if self.is_logging:
            button_text = "Disable Logging"

        self._toggle_logging_button.config(text=button_text)

    def toggle_autorefresh(self):
        autorefresh = self._autorefresh_var.get()
        if autorefresh:
            # self.send_message("Turn on logging auto-refresh")
            self.autorefresh_logging()
            self._refresh_button.configure(state='disabled', text="Disabled")
        else:
            # self.send_message("Turn off logging auto-refresh")
            self._refresh_button.configure(state='normal', text="Refresh")

    def toggle_log_to_terminal(self):
        log_to_terminal = self._log_to_terminal_var.get()
        if log_to_terminal:  # The order here is important so that the turn-on/off is always logged to the terminal
            self._logger.propagate = log_to_terminal
            # self.send_message("Turn on logging to terminal")
        else:
            # self.send_message("Turn off logging to terminal")
            self._logger.propagate = log_to_terminal

    def autorefresh_logging(self):
        self.refresh_logging()

        autorefresh = self._autorefresh_var.get()
        if autorefresh:
            self._text_display.after(500, self.autorefresh_logging)

    def refresh_logging(self):
        # pos = self._scrollbar.get()
        vw = self._text_display.yview()

        # print(pos)
        # print(vw)
        # TODO: Scrollbar is still jumping around when updating. It is related to when the lines of text wrap to the next line
        # Disabling line wrapping seems to have "fixed" (hidden) the issue

        self._text_display.configure(state='normal')
        self._text_display.delete("1.0", tk.END)
        self._text_display.insert('end', self.get_log())
        self._text_display.configure(state='disabled')
        # self._text_display.yview_moveto(pos[0])
        self._text_display.yview_moveto(vw[0])



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

__version__ = '0.1.9'

from .chips.etroc2_chip import ETROC2_Chip
from .functions import addLoggingLevel
from .i2c_messages import I2CMessages
from .i2c_usb_iss_helper import USB_ISS_Helper

# Add custom log levels to logging
addLoggingLevel('TRACE', 8)
addLoggingLevel('DETAILED_TRACE', 5)
# addLoggingLevel('HIGH_TEST', 100)

__all__ = ["I2CMessages", "USB_ISS_Helper", "ETROC2_Chip"]
